@@ -59,14 +59,29 @@ class BaseSettings(attrmap.AttrMap):
         return s
 
     @classmethod
-    def load_with_defaults(cls, file, check_args=True):
+    def load_with_defaults(cls, file, with_type_keyword=False, check_args=True):
         c_file = cls.load(file, check_args=check_args)
         c_default = cls.get_defaults()
 
         if 'default_settings' in c_file and c_file.default_settings:
             c_default += cls.load(c_file.get_path('default_settings'), check_args=check_args)
 
-        s = c_default + c_file
+        if not with_type_keyword:
+            s = c_default + c_file
+        else:
+            for k, v in c_file.items():
+                if isinstance(v, dict):
+                    if 'type' in v.keys():
+                        t = v['type']
+                    else:
+                        t = k
+                    if t not in c_default.keys():
+                        print(f'Warning: no default for section {k} of type {t}')
+                        continue
+                    t_default = c_default[t]
+                    c_file[k] = attrmap.AttrMap(t_default) + attrmap.AttrMap(v)
+            s = c_file
+
         s.set_file(file)
         return s
 
