@@ -40,20 +40,41 @@ def test_worker_log(tmp_path):
     assert len(success) == 20
     assert len(errors) == 0
 
-    for i in range(20):
+    for i, task in enumerate(success):
         assert os.path.isfile(tmp_path / f't{i}.log')
         with open(tmp_path / f't{i}.log') as f:
             s = f.readlines()
-        assert len(s) == 9 or len(s) == 7
+        assert len(s) == 10 or len(s) == 8
         assert s[0].startswith('# Logging starting at ')
         assert s[1] == f'# Input command: echo Task {i}\n'
+        if len(s) == 8:
+            assert s[2] == f'Starting on host {task.process.client.host} with PID {task.process.process.pid}\n'
+            assert s[4] == f'Task {i}\n'
+        if len(s) == 10:
+            assert s[2].startswith('Error')
+            assert s[3] == f'Will retry 1 / 3 ...\n'
+            assert s[4] == f'Starting on host {task.process.client.host} with PID {task.process.process.pid}\n'
+            assert s[6] == f'Task {i}\n'
         assert s[-1].startswith('# Logging stopped at ')
-        if len(s) == 7:
-            assert s[3] == f'Task {i}\n'
-        if len(s) == 9:
-            assert s[3].startswith('Error')
-            assert s[4] == f'Will retry 1 / 3 ...\n'
-            assert s[5] == f'Task {i}\n'
+
+
+def test_pidfile(tmp_path):
+    w = worker.WorkerPool([worker.localhost_shortname], 'Test')
+
+    for i in range(20):
+        w.add(f'echo Task {i}; exit {i}', pid_filename=f't{i}.pid', keep_pid_file_on_error=True)
+
+    success, errors = w.execute()
+
+    assert len(success) == 20
+    assert len(errors) == 0
+
+    for task in success:
+        assert os.path.exists(task.pid_filename) != (task.returncode == 0)
+        if task.returncode > 0:
+            with open(task.pid_filename) as fd:
+                s = fd.read()
+            assert s == f'{task.process.client.host} {task.process.process.pid}'
 
 
 def test_worker_wrong_server(tmp_path):
