@@ -4,6 +4,7 @@
 
 import os
 import re
+import sys
 import time
 import socket
 import signal
@@ -70,6 +71,7 @@ class WorkerStatus(Enum):
     COMMITED = 2
     RUNNING = 3
     ERROR = 4
+    CLOSED = 5
 
 
 class Process(object):
@@ -89,7 +91,8 @@ class Task(object):
         returncode (int): Return code. None if task was not executed.
     """
 
-    def __init__(self, name, command, output_file=None, done_callback=None, run_on_host=None, n_try=3):
+    def __init__(self, name, command, output_file=None, done_callback=None, 
+                 run_on_host=None, n_try=3, pid_filename=None, keep_pid_file_on_error=False):
         self.name = name
         self.command = command
         self.output_file = output_file
@@ -102,25 +105,40 @@ class Task(object):
         self.process = None
         self.n_try = n_try
         self.timed_out = False
+        self.pid_filename = pid_filename
+        self.keep_pid_file_on_error = keep_pid_file_on_error
 
     def init_log(self):
         if self.output_file is not None:
             self.fd = open(self.output_file, 'w')
             self.fd.write(f'# Logging starting at {datetime.datetime.now()}\n')
-            self.fd.write(f'# Input command: {self.command}\n\n')
+            self.fd.write(f'# Input command: {self.command}\n')
 
     def set_process(self, process):
         self.process = process
+        if hasattr(self.process.process, 'pid'):
+            self.log(f'Starting on host {self.process.client.host} with PID {self.process.process.pid}\n\n', 
+                     self.process.client.host)
+            if self.pid_filename is not None:
+                try:
+                    with open(self.pid_filename, 'w') as fd:
+                        fd.write(f'{self.process.client.host} {self.process.process.pid}')
+                    self.log(f'Starting on host {self.process.client.host} with PID {self.process.process.pid}\n\n',
+                             self.process.client.host)
+                except Exception:
+                    self.log(f'Error creating PID file {self.pid_filename}', self.process.client.host, err=True)
 
     async def terminate(self):
-        if self.process is not None:
-            if hasattr(self.process.process, 'pid'):
-                if self.process.client.host is not 'local':
-                    self.log(f'Killing local process PID {self.process.process.pid}\n', 'local', err=True)
+        if self.process is not None and hasattr(self.process.process, 'pid'):
+            if self.process.client.host is localhost_shortname:
+                self.log(f'Killing local process PID {self.process.process.pid}\n', 'local', err=True)
+                try:
                     os.killpg(os.getpgid(self.process.process.pid), signal.SIGTERM)
-                else:
-                    self.log(f'Killing remote process PID {self.process.pid}\n', self.process.client.host, err=True)
-                    kill_remote(self.process.client.host, self.process.client.user, self.process.pid)
+                except ProcessLookupError:
+                    print(f'Process {self.process.process.pid} not found')
+            else:
+                self.log(f'Killing remote process PID {self.process.process.pid}\n', self.process.client.host, err=True)
+                kill_remote(self.process.client.host, self.process.client.user, self.process.process.pid)
 
     def log(self, line, host, err=False):
         if isinstance(line, bytes):
@@ -143,9 +161,12 @@ class Task(object):
             break
 
     def close_log(self):
+        if self.pid_filename and os.path.exists(self.pid_filename):
+            if not self.keep_pid_file_on_error or self.returncode == 0:
+                os.remove(self.pid_filename)
         if self.fd is not None:
             try:
-                self.fd.write(f'\n\n# Logging stopped at {datetime.datetime.now()}\n')
+                self.fd.write(f'\n\n# Logging stopped at {datetime.datetime.now()} with code: {self.returncode}\n')
                 self.fd.close()
             except Exception:
                 print(f'Error closing log file for task {self.name}. Ignoring.')
@@ -223,9 +244,13 @@ class Client(object):
 
         async with self.creating_session:
             process = await self.conn.create_process('echo $$;' + task.command)
-        task.set_process(Process(self, process))
+        
+        try:
+            process.pid = int(await process.stdout.readline())
+        except ValueError:
+            task.log('Error retrieving PID of remote process', self.host, err=True)
 
-        process.pid = await process.stdout.readline()
+        task.set_process(Process(self, process))
 
         await task.log_stream(process.stdout, self.host)
         await task.log_stream(process.stderr, self.host, err=True)
@@ -309,6 +334,9 @@ class Worker(object):
             err_msg = f'Error connecting client: {str(exc)}\n'
             self.status = WorkerStatus.ERROR
             return False, err_msg
+        except asyncio.CancelledError:
+            err_msg = f'Task {self.name} has been canceled\n'
+            return True, err_msg
         except Exception as exc:
             err_msg = f'Error executing task {self.name}: {str(exc)}\n'
             self.status = WorkerStatus.IDLE
@@ -324,8 +352,9 @@ class Worker(object):
 
         return True, err_msg
 
-    def done(self):
-        self.status = WorkerStatus.IDLE
+    def close(self):
+        self.status = WorkerStatus.CLOSED
+        self.client.close()
 
 
 class WorkerPool(object):
@@ -357,7 +386,6 @@ class WorkerPool(object):
         self.dry_run = dry_run
         self.hosts = hosts
         self.futures = None
-        self.clients = []
         self.active_hosts = hosts
 
         for host in self.hosts:
@@ -393,7 +421,8 @@ class WorkerPool(object):
                     return worker
             await asyncio.sleep(0.1)
 
-    def add(self, command, name=None, output_file=None, done_callback=None, run_on_host=None, n_try=3):
+    def add(self, command, name=None, output_file=None, done_callback=None, 
+            run_on_host=None, n_try=3, pid_filename=None, keep_pid_file_on_error=False):
         """Add a command to execute by a worker in the pool. Optionally output result in output_file.
 
         Args:
@@ -415,7 +444,8 @@ class WorkerPool(object):
             return
 
         self.tasks.append(Task(name, command, output_file=output_file, done_callback=done_callback,
-                               run_on_host=run_on_host, n_try=n_try))
+                               run_on_host=run_on_host, n_try=n_try, pid_filename=pid_filename, 
+                               keep_pid_file_on_error=keep_pid_file_on_error))
 
     def get_all_tasks(self):
         tasks = collections.defaultdict(list)
@@ -450,12 +480,30 @@ class WorkerPool(object):
 
                 if n_active_workers == 0:
                     secho("No active workers, canceling all remaining tasks ...", fg='red')
-                    for future in self.futures[::-1]:
-                        future.cancel()
+                    if self.futures is not None:
+                        for future in self.futures[::-1]:
+                            future.cancel()
 
         except asyncio.CancelledError:
             pass
 
+    async def terminate_all(self):
+        for worker in self.workers:
+            worker.close()
+        for task in self.tasks:
+            if task.status == TaskStatus.RUNNING:
+                await task.terminate()
+        if self.futures is not None:
+            for task in self.futures:
+                task.cancel()
+
+    async def shutdown(self, exit_code=0):
+        await self.terminate_all()
+        await asyncio.sleep(0.5)
+
+        loop = asyncio.get_event_loop()
+        loop.call_soon_threadsafe(sys.exit, exit_code)
+
     async def process_tasks(self):
         self.pbar.start(max_value=len(self.tasks))
 
@@ -471,13 +519,18 @@ class WorkerPool(object):
 
         loop = asyncio.get_event_loop()
 
+        loop.add_signal_handler(signal.SIGTERM, lambda: asyncio.create_task(self.shutdown(exit_code=1)))
+
         try:
             loop.run_until_complete(self.process_tasks())
+        except KeyboardInterrupt:
+            loop.run_until_complete(self.terminate_all())
+            raise
         except asyncio.CancelledError:
             secho(f'Error: one (or more) task(s) have been canceled', fg='red')
         finally:
-            for client in self.clients:
-                client.close()
+            for worker in self.workers:
+                worker.close()
 
         tasks = self.get_all_tasks()
         n_success = len(tasks[TaskStatus.SUCCESS])
